#include <MFRC522v2.h>
#include <MFRC522DriverI2C.h>
#include <MFRC522Debug.h>

const uint8_t customAddress = 0x28;
TwoWire &customI2C = Wire;
MFRC522DriverI2C driver{customAddress, customI2C}; // Create I2C driver.
MFRC522 mfrc522{driver}; // Create MFRC522 instance.rnc

void setup() {
  Serial.begin(9600); // Initialize serial communications with the PC for debugging.
  while (!Serial);      // Do nothing if no serial port is opened (added for Arduinos based on ATMEGA32U4).

  mfrc522.PCD_Init();   // Init MFRC522 board.
  MFRC522Debug::PCD_DumpVersionToSerial(mfrc522, Serial);  // Show details of PCD - MFRC522 Card Reader details.
  Serial.println(F("Scan PICC to see UID, SAK, type, and data blocks..."));
}

void loop() {
  cardRead( );
//  if ( !mfrc522.PICC_IsNewCardPresent() || !mfrc522.PICC_ReadCardSerial()) {
//    return;
//  }
//
//  MFRC522Debug::PICC_DumpToSerial(mfrc522, Serial, &(mfrc522.uid));
}

//*****************************************************************************************//
void printHex(byte *buffer, byte bufferSize) {
  for (byte i = 0; i < bufferSize; i++) {
    Serial.print(buffer[i] < 0x10 ? " 0" : " ");
    Serial.print(buffer[i], HEX);
  }
}

void dispCardID(byte *buffer, byte bufferSize) {
  Serial.print(F("UID:"));
  printHex(buffer,bufferSize) ;
  Serial.println();
}

void dispReadError( ) {
  Serial.println(F("ERROR:所属不明"));
}

void cardError( ) {
  Serial.println(F("CARD ERROR"));
}

void dispCardData(char *buff1,char *buff2,char *buff3) {
  if (('0' <= buff1[0] <= '9') && 
      ('0' <= buff1[1] <= '9') &&
      ('0' <= buff1[2] <= '9') &&
      ('0' <= buff1[3] <= '9') &&
      buff1[4] == 0) {
    Serial.println(buff1);
    Serial.println(buff2);
    Serial.println(buff3);
  } else {
    dispReadError() ;
  }
}

//*****************************************************************************************//
void cardRead( ) {
  // Prepare key - all keys are set to FFFFFFFFFFFFh at chip delivery from the factory.
  MFRC522::MIFARE_Key key;
  for (byte i = 0; i < 6; i++) key.keyByte[i] = 0xFF;

  //some variables we need
  byte block;
  byte len;
  MFRC522::StatusCode status;

  //-------------------------------------------
  // Reset the loop if no new card present on the sensor/reader. This saves the entire process when idle.
  if ( ! mfrc522.PICC_IsNewCardPresent()) {
    return ;
  }

  // Select one of the cards
  if ( ! mfrc522.PICC_ReadCardSerial()) {
    return ;
  }
  
  // Check is the PICC of Classic MIFARE type
  MFRC522::PICC_Type piccType = mfrc522.PICC_GetType(mfrc522.uid.sak);
  if (piccType != MFRC522Constants::PICC_TYPE_MIFARE_MINI &&  
    piccType != MFRC522Constants::PICC_TYPE_MIFARE_1K &&
    piccType != MFRC522Constants::PICC_TYPE_MIFARE_4K) {
    dispCardID(mfrc522.uid.uidByte, mfrc522.uid.size);
    cardError( );
    Serial.println(F("Not MIFARE Classic."));
    return ;
  }

  //------------------------------------------- GET DATA
  byte buffer[32];
  byte buffer1[32];
  byte buffer2[32];
  byte buffer3[32];
  memset((char *)buffer1,0,32) ;
  memset((char *)buffer2,0,32) ;
  memset((char *)buffer3,0,32) ;

  //------------------------------------------- GET NO
  block = 4;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer1,buffer,16) ;

  //---------------------------------------- GET NAME E
  block = 5 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer2,buffer,16) ;

  block = 6 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(&buffer2[16],buffer,16) ;

  //---------------------------------------- GET NAME J
  block = 8 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer3,buffer,16) ;

  block = 9 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(&buffer3[16],buffer,16) ;

  mfrc522.PICC_HaltA();
  mfrc522.PCD_StopCrypto1();

  //----------------------------------------
  dispCardID(mfrc522.uid.uidByte, mfrc522.uid.size);
  dispCardData((char *)buffer1,(char *)buffer2,(char *)buffer3) ;
}
