#include <M5Stack.h>
#include <MFRC522v2.h>
#include <MFRC522DriverI2C.h>
#include <MFRC522Debug.h>
#include <SD.h>

#include "efont.h"               // Unicode表示用ライブラリ
#include "efontESP32.h"          // Unicode表示用ライブラリ　
#include "efontEnableJa.h"       // Unicode表示用ライブラリ（日本語）

const uint8_t customAddress = 0x28;
TwoWire &customI2C = Wire;
MFRC522DriverI2C driver{customAddress, customI2C}; // Create I2C driver.
MFRC522 mfrc522{driver}; // Create MFRC522 instance.rnc

char strBuff[80] ;
int selNo = 0 ;
int selNum = 6 ;
static struct Data_t {
  char IDNo[16] ;
  char NameE[32] ;
  char NameJ[32] ;
} data[] = { "0001","SASARA SATO","さとう ささら",
             "0002","TUZUMI SUDUKI","すずき つづみ",
             "0003","AMATO TAKAHASHI","タカハシ アマト",
             "0004","RIKKA KOHARU","小春 六花",
             "0005","KARIN NATUKI","夏色 花梨",
             "0006","CHIFUYU HANAKUMA","花隈 千冬"} ;

bool flagWriteMode = false ;

//*****************************************************************************************//
void setup() {
  M5.begin();        // Init M5Core.  初始化 M5Core

  M5.Power.begin();  // Init Power module.  初始化?源模?
  M5.Lcd.clear(BLACK);
  M5.Lcd.setTextFont(4);
  M5.Lcd.setTextSize(1);
  
  Serial.begin(9600);                    // Initialize serial communications with the PC
  while(!Serial) ;

  mfrc522.PCD_Init();                    // Init MFRC522 card
  dispMenu( ) ;
}

//*****************************************************************************************//
void loop() {
  M5.update();
  if (M5.BtnA.wasPressed()==1) {
    selNo = (selNo + 1) % selNum ;
    dispMenu( ) ;
    return ;
  }
  if (M5.BtnC.wasPressed()==1) {
    flagWriteMode = true ;
    dispMenu( ) ;
    return ;
  }
  if (flagWriteMode) {
    cardWrite(selNo) ;
  } else {
    cardRead( );
  }
}

//*****************************************************************************************//
void dispMenu( ) {
  M5.Lcd.setTextSize(1);
  M5.Lcd.clear(BLACK);
  M5.Lcd.setCursor(0,0);
  M5.Lcd.setTextColor(BLUE);
  M5.Lcd.print("DATA WRITE to REF CARD.");
  M5.Lcd.drawLine(0, 24, 319, 24 , WHITE);
  M5.Lcd.setCursor(30,210);
  M5.Lcd.setTextColor(GREEN);
  M5.Lcd.print("[SEL]");
  M5.Lcd.setCursor(200,210);
  M5.Lcd.setTextColor((flagWriteMode ? MAGENTA : DARKGREY));
  M5.Lcd.print("[WRITE]");

  M5.Lcd.drawLine(0, 166, 319, 166 , WHITE);
  M5.Lcd.setCursor(0,170);
  M5.Lcd.setTextColor(DARKGREY,BLACK);
  sprintf(strBuff,"%s %s",data[selNo].IDNo,data[selNo].NameJ) ;
  printEfont(strBuff,0,170,2);
}

void dispCardData(char *buff1,char *buff2,char *buff3) {
  M5.Lcd.setTextSize(1);
  M5.Lcd.setTextColor(WHITE,BLACK);
  M5.Lcd.setCursor(0,60);
  M5.Lcd.println(F("No:"));
  M5.Lcd.setCursor(0,95);
  M5.Lcd.println(F("E :"));
  M5.Lcd.setCursor(0,130);
  M5.Lcd.println(F("J :"));

  printEfont(buff1,45,50,2);
  printEfont(buff2,45,85,2);
  printEfont(buff3,45,120,2);
}

void printHex(byte *buffer, byte bufferSize) {
  for (byte i = 0; i < bufferSize; i++) {
    M5.Lcd.print(buffer[i] < 0x10 ? " 0" : " ");
    M5.Lcd.print(buffer[i], HEX);
  }
}

void printHexS(byte *buffer, byte bufferSize) {
  Serial.print("printHex:") ;
  for (byte i = 0; i < bufferSize; i++) {
    Serial.print(buffer[i] < 0x10 ? " 0" : " ");
    Serial.print(buffer[i], HEX);
  }
  Serial.println( );
}

void dispCardID(byte *buffer, byte bufferSize) {
  M5.Lcd.setTextSize(1);
  M5.Lcd.setTextColor(WHITE,BLACK);
  M5.Lcd.setCursor(0,26);
  M5.Lcd.print(F("ID:"));
  printHex(buffer,bufferSize) ;
  M5.Lcd.println();
}

//*****************************************************************************************//
void cardRead( ) {
  // Prepare key - all keys are set to FFFFFFFFFFFFh at chip delivery from the factory.
  MFRC522::MIFARE_Key key;
  for (byte i = 0; i < 6; i++) key.keyByte[i] = 0xFF;

  //some variables we need
  byte block;
  byte len;
  MFRC522::StatusCode status;

  //-------------------------------------------
  // Reset the loop if no new card present on the sensor/reader. This saves the entire process when idle.
  if ( ! mfrc522.PICC_IsNewCardPresent()) {
    return ;
  }

  // Select one of the cards
  if ( ! mfrc522.PICC_ReadCardSerial()) {
    return ;
  }
  
  Serial.println(F("**Card Detected:**"));
  M5.Lcd.fillRect(0, 25, 320, 141, BLACK);

  // Check is the PICC of Classic MIFARE type
  MFRC522::PICC_Type piccType = mfrc522.PICC_GetType(mfrc522.uid.sak);
  if (piccType != MFRC522Constants::PICC_TYPE_MIFARE_MINI &&  
    piccType != MFRC522Constants::PICC_TYPE_MIFARE_1K &&
    piccType != MFRC522Constants::PICC_TYPE_MIFARE_4K) {
    dispCardID(mfrc522.uid.uidByte, mfrc522.uid.size);
    M5.Lcd.println(F("Not MIFARE Classic."));
    delay(3000) ; // コレは表示の見栄えを良くするためのdelay、なくても良い。
    return ;
  }

  //------------------------------------------- GET DATA
  byte buffer[32];
  byte buffer1[32];
  byte buffer2[32];
  byte buffer3[32];
  memset((char *)buffer1,0,32) ;
  memset((char *)buffer2,0,32) ;
  memset((char *)buffer3,0,32) ;

  //------------------------------------------- GET NO
  block = 4;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("Authentication failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("1:Reading failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer1,buffer,16) ;

  //---------------------------------------- GET NAME E
  block = 5 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("Authentication failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("2:Reading failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer2,buffer,16) ;

  block = 6 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("Authentication failed: "));
    Serial.println(GetStatusCodeName(status));

    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("3:Reading failed: "));
    Serial.println(GetStatusCodeName(status));

    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(&buffer2[16],buffer,16) ;

  //---------------------------------------- GET NAME J
  block = 8 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("Authentication failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("2:Reading failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(buffer3,buffer,16) ;

  block = 9 ;
  len = 32;
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("Authentication failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("3:Reading failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return ;
  }
  memcpy(&buffer3[16],buffer,16) ;

  //----------------------------------------
  dispCardID(mfrc522.uid.uidByte, mfrc522.uid.size);
  printHexS(buffer1, sizeof(buffer1)) ;
  printHexS(buffer2, sizeof(buffer2)) ;
  printHexS(buffer3, sizeof(buffer3)) ;
  dispCardData((char *)buffer1,(char *)buffer2,(char *)buffer3) ;
  Serial.println(F("\n**End Reading**\n"));
  delay(1000); //change value if you want to read cards faster

  mfrc522.PICC_HaltA();
  mfrc522.PCD_StopCrypto1();
}

const char *GetStatusCodeName(MFRC522Constants::StatusCode code) {
  switch (code) {
    case MFRC522Constants::STATUS_OK:           return "Success.";
    case MFRC522Constants::STATUS_ERROR:        return "Error in communication.";
    case MFRC522Constants::STATUS_COLLISION:    return "Collision detected.";
    case MFRC522Constants::STATUS_TIMEOUT:      return "Timeout in communication.";
    case MFRC522Constants::STATUS_NO_ROOM:      return "A buffer is not big enough.";
    case MFRC522Constants::STATUS_INTERNAL_ERROR: return "Internal error in the code. Should not happen.";
    case MFRC522Constants::STATUS_INVALID:      return "Invalid argument.";
    case MFRC522Constants::STATUS_CRC_WRONG:    return "The CRC_A does not match.";
    case MFRC522Constants::STATUS_MIFARE_NACK:  return "A MIFARE PICC responded with NAK.";
    default:          return "Unknown error";
  }
} // End GetStatusCodeName()

const char *PICC_GetTypeName(MFRC522Constants::PICC_Type piccType) {
  switch (piccType) {
    case MFRC522Constants::PICC_TYPE_ISO_14443_4:   return "PICC compliant with ISO/IEC 14443-4";
    case MFRC522Constants::PICC_TYPE_ISO_18092:     return "PICC compliant with ISO/IEC 18092 (NFC)";
    case MFRC522Constants::PICC_TYPE_MIFARE_MINI:   return "MIFARE Mini, 320 bytes";
    case MFRC522Constants::PICC_TYPE_MIFARE_1K:     return "MIFARE 1KB";
    case MFRC522Constants::PICC_TYPE_MIFARE_4K:     return "MIFARE 4KB";
    case MFRC522Constants::PICC_TYPE_MIFARE_UL:     return "MIFARE Ultralight or Ultralight C";
    case MFRC522Constants::PICC_TYPE_MIFARE_PLUS:   return "MIFARE Plus";
    case MFRC522Constants::PICC_TYPE_MIFARE_DESFIRE:  return "MIFARE DESFire";
    case MFRC522Constants::PICC_TYPE_TNP3XXX:       return "MIFARE TNP3XXX";
    case MFRC522Constants::PICC_TYPE_NOT_COMPLETE:  return "SAK indicates UID is not complete.";
    case MFRC522Constants::PICC_TYPE_UNKNOWN:
    default:            return "Unknown type";
  }
} // End PICC_GetTypeName()

//*****************************************************************************************//
void cardWrite(int dataNO) {
  // Prepare key - all keys are set to FFFFFFFFFFFFh at chip delivery from the factory.
  MFRC522::MIFARE_Key key;
  for (byte i = 0; i < 6; i++) key.keyByte[i] = 0xFF;

  // Reset the loop if no new card present on the sensor/reader. This saves the entire process when idle.
  if ( ! mfrc522.PICC_IsNewCardPresent()) {
    return;
  }

  // Select one of the cards
  if ( ! mfrc522.PICC_ReadCardSerial()) {
    return;
  }

  M5.Lcd.fillRect(0, 25, 320, 141, BLACK);
  dispCardID(mfrc522.uid.uidByte, mfrc522.uid.size);
  M5.Lcd.setTextColor(RED);
  sprintf(strBuff,"DATA WRITE - No.%s",data[dataNO].IDNo) ;
  M5.Lcd.print(strBuff);

  Serial.print(F("Card UID:"));    //Dump UID
  for (byte i = 0; i < mfrc522.uid.size; i++) {
    Serial.print(mfrc522.uid.uidByte[i] < 0x10 ? " 0" : " ");
    Serial.print(mfrc522.uid.uidByte[i], HEX);
  }
  Serial.print(F(" PICC type: "));   // Dump PICC type

  MFRC522::PICC_Type piccType = mfrc522.PICC_GetType(mfrc522.uid.sak);
  Serial.println(PICC_GetTypeName(piccType));

//  byte buffer[34];
  byte *buffer ;
  byte block;
  MFRC522::StatusCode status;
  byte len;

//  Serial.setTimeout(20000L) ;     // wait until 20 seconds for input from serial
//  // Ask personal data: Family name
//  Serial.println(F("Type ID No, ending with #"));
//  len = Serial.readBytesUntil('#', (char *) buffer, 16) ; // read family name from serial
//  for (byte i = len; i < 16; i++) buffer[i] = '\0';     // pad with spaces

  buffer = (byte *)data[dataNO].IDNo ;
  block = 4;
  //Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("PCD_Authenticate() success: "));

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer, 16);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("MIFARE_Write() success: "));

//  // Ask personal data: First name
//  Serial.println(F("Type NAME E, ending with #"));
//  len = Serial.readBytesUntil('#', (char *) buffer, 32) ; // read first name from serial
//  for (byte i = len; i < 32; i++) buffer[i] = '\0';     // pad with spaces

  buffer = (byte *)data[dataNO].NameE ;
  block = 5;
  //Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer, 16);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("MIFARE_Write() success: "));

  block = 6;
  //Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, &buffer[16], 16);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("MIFARE_Write() success: "));

//  // Ask personal data: First name
//  Serial.println(F("Type NAME J, ending with #"));
//  len = Serial.readBytesUntil('#', (char *) buffer, 32) ; // read first name from serial
//  for (byte i = len; i < 32; i++) buffer[i] = '\0';     // pad with spaces

  buffer = (byte *)data[dataNO].NameJ;
  block = 8;
  //Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer, 16);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("MIFARE_Write() success: "));

  block = 9;
  //Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, &buffer[16], 16);
  if (status != MFRC522Constants::STATUS_OK) {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(GetStatusCodeName(status));
    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
    return;
  }
  else Serial.println(F("MIFARE_Write() success: "));


  Serial.println(" ");
  mfrc522.PICC_HaltA(); // Halt PICC
  mfrc522.PCD_StopCrypto1();  // Stop encryption on PCD

  flagWriteMode = false ;
  dispMenu( ) ;
}
//*****************************************************************************************//
